"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List, Set,
    Sized,
    Tuple,
    Type,
    TypeVar,
    Union,
)


GHOST_PREFIX = "_gh_"

CONTRACT_WRAPPER_FUNCS = ['Requires', 'Ensures', 'Exsures', 'Invariant']

CONTRACT_FUNCS = ['Assume', 'Assert', 'Old', 'Result', 'Implies', 'Forall',
                  'Exists', 'Low', 'LowVal', 'LowEvent', 'Declassify', 'TerminatesSif',
                  'Acc', 'Rd', 'Wildcard', 'Fold', 'Unfold', 'Unfolding', 'Previous',
                  'RaisedException', 'PSeq', 'PSet', 'ToSeq', 'MaySet', 'MayCreate',
                  'getMethod', 'getArg', 'getOld', 'arg', 'Joinable', 'MayStart', 'Let',
                  'PMultiset',]

T = TypeVar('T')
V = TypeVar('V')


def Requires(expr: bool) -> bool:
    pass


def Ensures(expr: bool) -> bool:
    pass


def Exsures(exception: type, expr: bool) -> bool:
    pass


def Invariant(expr: bool) -> bool:
    pass


def Assume(expr: bool) -> None:
    pass


def Assert(expr: bool) -> bool:
    pass


def Old(expr: T) -> T:
    pass


def Result() -> Any:
    pass


def RaisedException() -> Any:
    pass


def Implies(p: bool, q: bool) -> bool:
    """
    Logical implication p ==> q.
    """
    pass


def Let(e1: T, t: Type[V], e2: Callable[[T], V]) -> V:
    """
    Allows defining an alias for a (pure) expression e1 to use in
    another expression or assertion e2.
    Let(5, int, lambda x : x + 34) means let x = 5 in x + 34
    """
    pass

def Forall(domain: Union[Iterable[T], Type[T]],
           predicate: Callable[[T], Union[bool, Tuple[bool, List[List[Any]]]]]) -> bool:
    """
    forall x in domain: predicate(x)
    """
    pass


def Exists(domain: Iterable[T], predicate: Callable[[T], bool]) -> bool:
    """
    exists x in domain: predicate(x)
    """
    pass


def Low(expr: T) -> bool:
    """
    Predicate to indicate that an expression has to be *low*.
    Ignored when not verifying information flow.
    """
    pass

def LowVal(expr: T) -> bool:
    """
    Predicate to indicate that an expression has to be low, using value equality if the
    expression is a primitive. Ignored when not verifying information flow.
    """
    pass

def LowEvent() -> bool:
    """
    Predicate that states that either both executions reach this point or none of them.
    """
    pass

def Declassify(expr: T) -> bool:
    """
    Declassify an expression. Assumes expression to be low.
    """
    pass

def TerminatesSif(cond: bool, rank: int) -> bool:
    """
    Verify absence of termination channels. Gives surrounding loop/call a
    termination condition and a ranking function.
    """
    pass

class PSeq(Generic[T], Sized, Iterable[T]):
    """
    A PSeq[T] represents a pure sequence of instances of subtypes of T, and
    is translated to native Viper sequences.
    """

    def __init__(self, *args: T) -> None:
        """
        ``PSeq(a, b, c)`` creates a PSeq instance containing the objects
        a, b and c in that order.
        """

    def __contains__(self, item: object) -> bool:
        """
        True iff this PSeq contains the given object (not taking ``__eq__``
        into account).
        """

    def __getitem__(self, item: int) -> T:
        """
        Returns the item at the given position.
        """

    def __len__(self) -> int:
        """
        Returns the length of this PSeq.
        """

    def __add__(self, other: 'PSeq[T]') -> 'PSeq[T]':
        """
        Concatenates two PSeqs of the same type to get a new PSeq.
        """

    def take(self, until: int) -> 'PSeq[T]':
        """
        Returns a new PSeq of the same type containing all elements starting
        from the beginning until the given index. ``PSeq(3,2,5,6).take(3)``
        is equal to ``PSeq(3,2,5)``.
        """

    def drop(self, until: int) -> 'PSeq[T]':
        """
        Returns a new PSeq of the same type containing all elements starting
        from the given index (i.e., drops all elements until that index).
        ``PSeq(2,3,5,6).drop(2)`` is equal to ``PSeq(5,6)``.
        """

    def update(self, index: int, new_val: T) -> 'PSeq[T]':
        """
        Returns a new sequence of the same type, containing the same elements
        except for the element at index ``index``, which is replaced by
        ``new_val``.
        """

    def __iter__(self) -> Iterator[T]:
        """
        PSeqs can be quantified over; this is only here so that PSeqs
        can be used as arguments for Forall.
        """

def Previous(it: T) -> PSeq[T]:
    """
    Within the body of a loop 'for x in xs', Previous(x) represents the list of
    the values of x in previous loop iterations.
    """
    pass


class PSet(Generic[T], Sized, Iterable[T]):
    """
    A PSet[T] represents a pure set of instances of subtypes of T, and is translated to
    native Viper sets.
    """

    def __init__(self, *args: T) -> None:
        """
        ``PSet(a, b, c)`` creates a set instance containing the objects
        a, b and c.
        """

    def __contains__(self, item: object) -> bool:
        """
        True iff this set contains the given object (not taking ``__eq__``
        into account).
        """

    def __len__(self) -> int:
        """
        Returns the cardinality of this set.
        """

    def __add__(self, other: 'PSet[T]') -> 'PSet[T]':
        """
        Returns the union of this set and the other.
        """

    def __sub__(self, other: 'PSet[T]') -> 'PSet[T]':
        """
        Returns the difference between this set and the other,
        """

    def __iter__(self) -> Iterator[T]:
        """
        Sets can be quantified over; this is only here so that sets
        can be used as arguments for Forall.
        """

class PMultiset(Generic[T], Sized, Iterable[T]):
    """
    An PMultiset[T] represents a pure multiset of instances of subtypes of T, and is translated to
    native Viper multisets.
    """

    def __init__(self, *args: T) -> None:
        """
        ``PMultiset(a, b, c)`` creates a multiset instance containing the objects
        a, b and c.
        """

    def num(self, item: object) -> int:
        """
        Returns the number of occurrences of ``item`` in this multiset.
        """

    def __len__(self) -> int:
        """
        Returns the cardinality of this set.
        """

    def __add__(self, other: 'PMultiset[T]') -> 'PMultiset[T]':
        """
        Returns the union of this multiset and the other.
        """

    def __sub__(self, other: 'PMultiset[T]') -> 'PMultiset[T]':
        """
        Returns the difference between this multiset and the other,
        """

    def __iter__(self) -> Iterator[T]:
        """
        Multisets can be quantified over; this is only here so that multisets
        can be used as arguments for Forall.
        """


def ToSeq(l: Iterable[T]) -> PSeq[T]:
    """
    Converts the given iterable of a built-in type (list, set, dict, range) to
    a pure PSeq.
    """


# The following annotations have no runtime semantics. They are only used for
# the Python to Viper translation.

def Acc(field, ratio=1) -> bool:
    """
    Access permission to field.
    0 < ratio < 1 means read-only access.
    ratio == 1 mean read-write access.
    """
    pass


def MayCreate(o: object, field_name: str) -> bool:
    """
    Permission to create a field called field_name on object o.
    """
    pass

def MaySet(o: object, field_name: str) -> bool:
    """
    Permission to either create a field called field_name on object o or access the
    existing field with that name.
    """
    pass


def Rd(field) -> bool:
    """
    Read permission to a predicate or field, only to be used in pure contexts.
    """
    pass


def ARP(counting: int = None) -> float:
    """
    Abstract read permission, only to be used in Acc(f, ...).
    """
    pass


"""
Permission used in predicates
"""
RD_PRED = 1  # type: float


def Wildcard(field) -> bool:
    """
    Wildcard permission to a predicate or field, only to be used in pure contexts.
    """
    pass


def Fold(predicate: bool) -> None:
    pass


def Unfold(predicate: bool) -> None:
    pass


def Unfolding(predicate: bool, expr: T) -> T:
    """
    Evaluates expr in a state where predicate has been unfolded.
    """
    return expr


def Pure(func: T) -> T:
    """
    Decorator to mark pure functions. It's a no-op.
    """
    return func


def Predicate(func: T) -> T:
    """
    Decorator to mark predicate functions. It's a no-op.
    """
    return func


def Ghost(func: T) -> T:
    """
    Decorator for ghost functions. It's a no-op.
    """
    return func

def AllLow(func: T) -> T:
    """
    Decorator indicating that everything this method does is low.
    Requires all inputs to be low, ensures all state it has access to and
    all return values are low.
    """
    return func

def PreservesLow(func: T) -> T:
    """
    Decorator indicating that everything this method does preserves lowness.
    Given that all the state it gets to work on is low to begin with, all state and
    return values will remain low.
    """
    return func

def ContractOnly(func: T) -> T:
    """
    Decorator to mark contract only methods. It's a no-op.
    """
    return func


def GhostReturns(start_index: int) -> Callable[[T], T]:
    """
    Decorator for functions which specifies which return values are ghost
    returns, starting at index 0. It's a no-op.
    If a function returns an n-tuple, @GhostReturns(k) means that
    elements 0 to k-1 are normal return values, elements k to n-1 are ghost
    return values. k must be less than n. If the function returns a value
    that is not a tuple, start_index can only be 0 (meaning that the only value
    that is returned is a ghost value). Using this decorator on functions which
    do not return anything is not allowed.
    """
    def wrap(func: T) -> T:
        return func
    return wrap


def list_pred(l: object) -> bool:
    """
    Special, predefined predicate that represents the permissions belonging
    to a list. To be used like normal predicates, except it does not need to
    be folded or unfolded.
    """
    pass


def set_pred(s: object) -> bool:
    """
    Special, predefined predicate that represents the permissions belonging
    to a set. To be used like normal predicates, except it does not need to
    be folded or unfolded.
    """
    pass


def dict_pred(d: object) -> bool:
    """
    Special, predefined predicate that represents the permissions belonging
    to a dict. To be used like normal predicates, except it does not need to
    be folded or unfolded.
    """


__all__ = [
        'Requires',
        'Ensures',
        'Exsures',
        'Invariant',
        'Previous',
        'Assume',
        'Assert',
        'Old',
        'Result',
        'RaisedException',
        'Implies',
        'Forall',
        'Exists',
        'Let',
        'Low',
        'LowVal',
        'LowEvent',
        'Declassify',
        'TerminatesSif',
        'AllLow',
        'PreservesLow',
        'Acc',
        'Rd',
        'ARP',
        'RD_PRED',
        'Wildcard',
        'Fold',
        'Unfold',
        'Unfolding',
        'Pure',
        'Predicate',
        'Ghost',
        'ContractOnly',
        'GhostReturns',
        'list_pred',
        'dict_pred',
        'set_pred',
        'PSeq',
        'PSet',
        'PMultiset',
        'ToSeq',
        'MaySet',
        'MayCreate',
        ]



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from nagini_translation.lib.constants import (
    BOOL_TYPE,
    FLOAT_TYPE,
    INT_TYPE,
    PSET_TYPE,
    PSEQ_TYPE,
    STRING_TYPE,
    TUPLE_TYPE
)
from nagini_translation.lib.program_nodes import MethodType
from nagini_translation.lib.typedefs import StmtsAndExpr, Info
from nagini_translation.lib.util import (
    InvalidProgramException,
    UnsupportedException,
)
from nagini_translation.sif.lib.util import (
    in_override_check,
    in_postcondition_of_dyn_bound_call,
)
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.contract import ContractTranslator


class SIFContractTranslator(ContractTranslator):
    """
    Extended AST version of the contract translator.
    """
    def _create_dyn_check_info(self, ctx: Context) -> Info:
        """
        Check if we are in the postcondition of a method which calls will be dynamically
        bound, if so create a SIFDynCheckInfo, else NoInfo.
        If we are in a override check method, only do the version with dynamic call check.
        """
        self_type = in_postcondition_of_dyn_bound_call(self.type_factory, ctx)
        if self_type:
            return self.viper.SIFDynCheckInfo([], self_type,
                                              dyn_check_only=in_override_check(ctx))
        return self.no_info(ctx)


    def translate_low(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the Low() contract function.
        """
        if len(node.args) != 1:
            raise UnsupportedException(node, "Low() requires exactly one argument")
        stmts, expr = self.translate_expr(node.args[0], ctx)
        if stmts:
            raise InvalidProgramException(node, 'purity.violated')
        info = self._create_dyn_check_info(ctx)
        return [], self.viper.Low(
            expr, None, self.to_position(node, ctx), info)

    def translate_lowval(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the Low() contract function.
        """
        stmts, expr = self.translate_expr(node.args[0], ctx)
        if stmts:
            raise InvalidProgramException(node, 'purity.violated')
        info = self._create_dyn_check_info(ctx)
        # determine the comparator function to use
        expr_type = self.get_type(node.args[0], ctx)
        low_val_types = [BOOL_TYPE, FLOAT_TYPE, INT_TYPE, PSET_TYPE, PSEQ_TYPE,
                         STRING_TYPE, TUPLE_TYPE]
        if expr_type.name in low_val_types:
            comparator = expr_type.get_function('__eq__')
            comparator = comparator.sil_name
        else:
            comparator = None
        return [], self.viper.Low(
            expr, comparator, self.to_position(node, ctx), info)

    def translate_lowevent(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the LowEvent() contract function.
        """
        if ctx.current_class and ctx.current_function.method_type == MethodType.normal:
            self_type = self.type_factory.typeof(
                next(iter(ctx.actual_function.args.values())).ref(), ctx)
            info = self.viper.SIFDynCheckInfo([], self_type,
                                              dyn_check_only=in_override_check(ctx))
        else:
            info = self.no_info(ctx)
        return [], self.viper.LowEvent(self.to_position(node, ctx), info)

    def translate_declassify(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the Declassify() contract function.
        """
        stmts, expr = self.translate_expr(node.args[0], ctx)
        if stmts:
            raise InvalidProgramException(node, 'purity.violated')
        return [self.viper.Declassify(
            expr, self.to_position(node, ctx), self.no_info(ctx))], None

    def translate_terminates_sif(self, node: ast.Call, ctx: Context) -> StmtsAndExpr:
        """
        Translates a call to the TerminatesSif() contract function.
        """
        cond_stmts, cond = self.translate_expr(node.args[0], ctx,
                                               target_type=self.viper.Bool)
        if cond_stmts:
            raise InvalidProgramException(node, 'purity.violated')
        must_terminate = super().translate_terminates_sif(node, ctx)
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        implication = self.viper.Implies(cond, must_terminate[1], pos, info)
        terminates_exp = self.viper.TerminatesSif(cond, pos, info)
        return must_terminate[0], self.viper.And(implication, terminates_exp, pos, info)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Conversion of errors to human readable messages."""

import ast
from nagini_translation.lib.util import (
    get_containing_member,
    get_target_name,
    pprint,
)


ERRORS = {
    'assignment.failed':
        lambda n: 'Assignment might fail.',
    'call.failed':
        lambda n: 'Method call might fail.',
    'not.wellformed':
        lambda n: 'Contract might not be well-formed.',
    'call.precondition':
        lambda n: ('The precondition of method {} might not '
                   'hold.'.format(get_target_name(n))
                   if isinstance(n, (ast.Call, ast.FunctionDef)) else
                   'The precondition of {} might not hold.'.format(pprint(n))),
    'application.precondition':
        lambda n: ('The precondition of function {} might not '
                   'hold.'.format(get_target_name(n))
                   if isinstance(n, (ast.Call, ast.FunctionDef)) else
                   'The precondition of {} might not hold.'.format(pprint(n))),
    'exhale.failed':
        lambda n: 'Exhale might fail.',
    'inhale.failed':
        lambda n: 'Inhale might fail.',
    'if.failed':
        lambda n: 'Conditional statement might fail.',
    'while.failed':
        lambda n: 'While statement might fail.',
    'assert.failed':
        lambda n: 'Assert might fail.',
    'postcondition.violated':
        lambda n: ('Postcondition of {} might not '
                   'hold.').format(get_containing_member(n).name),
    'fold.failed':
        lambda n: 'Fold might fail.',
    'unfold.failed':
        lambda n: 'Unfold might fail.',
    'invariant.not.preserved':
        lambda n: 'Loop invariant might not be preserved.',
    'invariant.not.established':
        lambda n: 'Loop invariant might not hold on entry.',
    'function.not.wellformed':
        lambda n: ('Function {} might not be '
                   'well-formed.').format(get_containing_member(n).name),
    'predicate.not.wellformed':
        lambda n: ('Predicate {} might not be '
                   'well-formed.').format(get_containing_member(n).name),
    'termination_check.failed':
        lambda n: 'Operation might not terminate.',
    'leak_check.failed':
        lambda n: 'Obligation leak check failed.',
    'internal':
        lambda n: 'An internal error occurred.',
    'expression.undefined':
        lambda n: 'Expression {} may not be defined.'.format(pprint(n)),
    'thread.creation.failed':
        lambda n: 'Thread creation may fail.',
    'thread.start.failed':
        lambda n: 'Thread start may fail.',
    'thread.join.failed':
        lambda n: 'Thread join may fail.',
    'termination_channel_check.failed':
        lambda n: 'Termination channel might exist.',
    'lock.invariant.not.established':
        lambda n: 'Lock invariant might not hold.',
}

REASONS = {
    'assertion.false':
        lambda n: 'Assertion {} might not hold.'.format(pprint(n)),
    'receiver.null':
        lambda n: 'Receiver of {} might be null.'.format(pprint(n)),
    'division.by.zero':
        lambda n: 'Divisor {} might be zero.'.format(pprint(n)),
    'negative.permission':
        lambda n: 'Fraction {} might be negative.'.format(pprint(n)),
    'insufficient.permission':
        lambda n: ('There might be insufficient permission to '
                   'access {}.').format(pprint(n)),
    'termination_measure.non_positive':
        lambda n: ('Termination measure {} might be '
                   'non-positive.').format(pprint(n)),
    'measure.non_decreasing':
        lambda n: ('Termination measure of {} might be not '
                   'smaller.').format(pprint(n)),
    'gap.enabled':
        lambda n: ('Gap {} might be enabled in terminating IO '
                   'operation.').format(pprint(n)),
    'child_termination.not_implied':
        lambda n: ('Parent IO operation termination condition does not '
                   'imply {} termination condition.').format(pprint(n)),
    'obligation_measure.non_positive':
        lambda n: ('Obligation {} measure might be '
                   'non-positive.').format(pprint(n)),
    'must_terminate.not_taken':
        lambda n: ('Callee {} might not take MustTerminate '
                   'obligation.').format(get_target_name(n)),
    'must_terminate.loop_not_promised':
        lambda n: ('Loop might not promise to terminate.'),
    'must_terminate.loop_promise_not_kept':
        lambda n: ('Loop might not keep promise to terminate.'),
    'caller.has_unsatisfied_obligations':
        lambda n: ('Callee {} might not take all unsatisfied obligations '
                   'from the caller.'.format(get_target_name(n))),
    'method_body.leaks_obligations':
        lambda n: ('Body of method {} might leak '
                   'obligations.'.format(get_target_name(n))),
    'loop_context.has_unsatisfied_obligations':
        lambda n: ('Loop might not take all unsatisfied obligations '
                   'from the context.'),
    'loop_body.leaks_obligations':
        lambda n: ('Loop body might leak obligations.'),
    'loop_condition.not_framed_for_obligation_use':
        lambda n: ('Loop condition part {} is not framed at the point where '
                   'obligation is used.'.format(pprint(n))),
    'undefined.local.variable':
        lambda n: 'Local variable may not have been defined.',
    'undefined.global.name':
        lambda n: 'Global name may not have been defined.',
    'missing.dependencies':
        lambda n: 'Global dependencies may not be defined.',
    'internal':
        lambda n: 'Internal Viper error.',
    'receiver.not.injective':
        lambda n: 'Receiver expression of quantified permission is not injective.',
    'wait.level.invalid':
        lambda n: 'Thread level may not be lower than current thread.',
    'thread.not.joinable':
        lambda n: 'Thread may not be joinable.',
    'invalid.argument.type':
        lambda n: 'Thread argument may not fit target method parameter type.',
    'method.not.listed':
        lambda n: "Thread's target method may not be listed in start statement.",
    'missing.start.permission':
        lambda n: 'May not have permission to start thread.',
    'sif.fold':
        lambda n: 'The low parts of predicate {} might not hold.'.format(
            get_target_name(n.args[0])),
    'sif.unfold':
        lambda n: 'The low parts of predicate {} might not hold.'.format(
            get_target_name(n.args[0])),
    'sif_termination.condition_not_low':
        lambda n: 'Termination condition {} might not be low.'.format(pprint(n.args[0])),
    'sif_termination.not_lowevent':
        lambda n: ('Termination condition {} evaluating to false might not imply that '
                   'both executions don\'t terminate.').format(pprint(n.args[0])),
    'sif_termination.condition_not_tight':
        lambda n: 'Termination condition {} might not be tight.'.format(pprint(n.args[0])),
}

VAGUE_REASONS = {
    'assertion.false': '',
    'receiver.null': 'Receiver might be null.',
    'division.by.zero': 'Divisor might be zero.',
    'negative.permission': 'Fraction might be negative.',
    'insufficient.permission': 'There might be insufficient permission.',
    'termination_measure.non_positive': 'Termination measure might be non-positive.',
    'measure.non_decreasing': 'Termination measure might not be smaller.',
    'gap.enabled': 'Gap might be enabled in terminating IO operation.',
    'child_termination.not_implied': ('Parent IO operation termination condition does '
                                      'not imply termination condition.'),
    'obligation_measure.non_positive': 'Obligation measure might be non-positive.',
    'must_terminate.not_taken': 'Callee might not take MustTerminate obligation.',
    'caller.has_unsatisfied_obligations': ('Callee might not take all unsatisfied '
                                           'obligations from the caller.'),
    'method_body.leaks_obligations': 'Method body might leak obligations.',
    'loop_condition.not_framed_for_obligation_use': ('Loop condition part is not framed '
                                                     'at the point where obligation is '
                                                     'used.'),
}



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Comparisons of integer typed Silver expressions."""


import abc

from nagini_translation.lib.silver_nodes.bool_expr import BoolExpression
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Position,
)


class IntComparison(BoolExpression):   # pylint: disable=abstract-method
    """A base class for all integer comparison operators.

    Except ``!=`` and ``==``, which are defined for all expressions.
    """

    def __init__(self, left: 'IntExpression', right: 'IntExpression') -> None:
        self._left = left
        self._right = right

    @abc.abstractmethod
    def _compare(self, left: int, right: int) -> bool:
        """Compare ``left`` to ``right``.

        That is, if the value is known statically, it evaluates the
        expression.
        """

    def is_always_true(self) -> bool:
        """Check if it is statically known that comparison is ``True``."""
        left_value = self._left.get_value()
        if left_value is None:
            return False
        right_value = self._right.get_value()
        if right_value is None:
            return False
        return self._compare(left_value, right_value)

    def is_always_false(self) -> bool:
        """Check if it is statically known that comparison is ``False``."""
        left_value = self._left.get_value()
        if left_value is None:
            return False
        right_value = self._right.get_value()
        if right_value is None:
            return False
        return not self._compare(left_value, right_value)


class GtCmp(IntComparison):
    """Greater than comparison."""

    def _compare(self, left: int, right: int) -> bool:
        return left > right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.GtCmp(
            left, right, position, info)


class GeCmp(IntComparison):
    """Greater equal comparison."""

    def _compare(self, left: int, right: int) -> bool:
        return left >= right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.GeCmp(
            left, right, position, info)


class LtCmp(IntComparison):
    """Less than comparison."""

    def _compare(self, left: int, right: int) -> bool:
        return left < right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.LtCmp(
            left, right, position, info)


class LeCmp(IntComparison):
    """Less equal comparison."""

    def _compare(self, left: int, right: int) -> bool:
        return left <= right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.LeCmp(
            left, right, position, info)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""The lock class stub."""

from typing import Generic, Optional, TypeVar

from nagini_contracts.contracts import (
    ContractOnly,
    Ensures,
    Implies,
    Low,
    LowEvent,
    Predicate,
    Pure,
    Requires,
)
from nagini_contracts.obligations import (
    BaseLock,
    Level,
    MustRelease,
    MustTerminate,
    WaitLevel,
)


T = TypeVar("T")


class Lock(BaseLock, Generic[T]):
    """A stub for a lock that protects (parts of) the state of an object of type T."""

    def __init__(self, locked_object: T,
                 above: Optional[BaseLock]=None,
                 below: Optional[BaseLock]=None) -> None:
        """
        Create a lock whose level is below that of ``below`` and above that of ``above``,
        which protects ``locked_object``.
        Creating the lock "shares" the object (i.e., exhales the invariant).
        Create subclasses of this class and override ``invariant`` to create a lock
        with an invariant.

        ``Level(above)`` defaults to ``WaitLevel()``.
        """
        Requires(MustTerminate(1))
        Requires(Implies(above is None and below is not None,
                         WaitLevel() < Level(below)))
        Requires(Implies(above is not None and below is not None,
                         Level(above) < Level(below)))
        Requires(self.invariant())
        Ensures(Implies(above is None, WaitLevel() < Level(self)))
        Ensures(Implies(above is not None, Level(above) < Level(self)))
        Ensures(Implies(below is not None, Level(self) < Level(below)))

    @Pure
    @ContractOnly
    def get_locked(self) -> T:
        """Returns the object protected by this lock."""

    @Predicate
    def invariant(self) -> bool:
        """
        The lock invariant, expressed in terms of ``self.get_locked()``. Override this
        in a subclass to create a lock with an invariant.
        """
        return True

    @ContractOnly
    def acquire(self) -> None:
        """Acquire the lock."""
        Requires(MustTerminate(1))
        Requires(WaitLevel() < Level(self))
        Requires(Low(self))
        Requires(LowEvent())
        Ensures(self.invariant())
        Ensures(MustRelease(self))

    @ContractOnly
    def release(self) -> None:
        """Release the lock."""
        Requires(MustTerminate(1))
        Requires(MustRelease(self, 1))
        Requires(self.invariant())
        Requires(Low(self))
        Requires(LowEvent())



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from typing import List, Tuple

from nagini_translation.translators.call import CallTranslator
from nagini_translation.translators.abstract import Context
from nagini_translation.lib.program_nodes import PythonMethod, PythonVar
from nagini_translation.lib.typedefs import Stmt

class SIFCallTranslator(CallTranslator):
    """
    Extended AST version of call translator.
    """

    def get_error_var(self, stmt: ast.AST, ctx: Context) -> 'silver.ast.LocalVarRef':
        return ctx.current_function.error_var.ref()

    def inline_method(self, method: PythonMethod, args: List[PythonVar],
                      result_var: PythonVar, error_var: PythonVar,
                      ctx: Context) -> Tuple[List[Stmt], 'silver.ast.Label']:
        stmts, _ = super().inline_method(method, args, result_var, error_var, ctx)
        pos = self.no_position(ctx)
        info = self.no_info(ctx)
        seqn = self.viper.Seqn(stmts, pos, info)
        return [self.viper.InlinedCall(seqn, pos, info)], None



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

# The ADT module provides support for specifying algebraic data types for Nagini.
# ADTs are meant to be used exclusively in contracts: although they are syntatically
# correct Python constructs, they should be defined with no functional intent while
# running the code. Still, Nagini will semantically consider it while verifying the
# code.
#
# Example: linked list
#
# Pseudo-code:
#
#    LinkedList = Null
#               | Node(elem: int, next: LinkedList)
#
# Python definition: 
#
#   class LinkedList(ADT):
#       pass
#   
#   class Node(LinkedList, NamedTuple('Node', [('elem', int), ('next', LinkedList)])):
#       pass
#   
#   class Null(LinkedList, NamedTuple('Null', [])):
#       pass
#
# Uses:
#
#   list_1 = Null()                         []
#   list_2 = Node(2, Null())                [2]
#   list_3 = Node(3, Node(4, Null()))       [3, 4]
#

class ADT:
    pass



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Boolean typed Silver expressions."""


import ast

from typing import List

from nagini_translation.lib.constants import PRIMITIVE_BOOL_TYPE
from nagini_translation.lib.program_nodes import (
    PythonVar,
)
from nagini_translation.lib.silver_nodes.call import CallMixin
from nagini_translation.lib.silver_nodes.expression import Expression
from nagini_translation.lib.silver_nodes.program import Field
from nagini_translation.lib.silver_nodes.types import BOOL
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Position,
)
from nagini_translation.lib.util import (
    join_expressions,
)


class BoolExpression(Expression):   # pylint: disable=abstract-method
    """A base class for all boolean expressions."""

    def is_always_true(self) -> bool:
        """Return if this expression translation would always yield True."""
        return False

    def is_always_false(self) -> bool:
        """Return if this expression translation would always yield False."""
        return False


class TrueLit(BoolExpression):
    """``True`` literal."""

    def is_always_true(self) -> bool:
        """Return if this expression translation would always yield True."""
        return True

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        return translator.viper.TrueLit(position, info)


class FalseLit(BoolExpression):
    """``False`` literal."""

    def is_always_false(self) -> bool:
        """Return if this expression translation would always yield False."""
        return True

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        return translator.viper.FalseLit(position, info)


class Not(BoolExpression):
    """Negation of other expression."""

    def __init__(self, value: BoolExpression) -> None:
        self._value = value

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        value = self._value.translate(translator, ctx, position, info)
        return translator.viper.Not(value, position, info)


class ForPerm(BoolExpression):
    """ForPerm expression."""

    def __init__(
            self, var_name: str, target: 'Predicate',
            body: BoolExpression) -> None:
        self._var_name = var_name
        self._target = target
        self._body = body

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        var = translator.viper.LocalVarDecl(
            self._var_name, translator.viper.Ref, position, info)
        if isinstance(self._target, Field):
            target = self._target.translate(translator, ctx, position, info)
            access = translator.viper.FieldAccess(var.localVar(), target, position, info)
        else:
            access = translator.viper.PredicateAccess([var.localVar()],
                                                      self._target._name, position, info)
        body = self._body.translate(translator, ctx, position, info)
        return translator.viper.ForPerm(
            var, access, body, position, info)


class PythonBoolExpression(BoolExpression):
    """An boolean expression represented by Python bool expression."""

    def __init__(self, node: ast.expr) -> None:
        self._node = node

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        stmt, expr = translator.translate_expr(
            self._node, ctx, target_type=translator.viper.Bool)
        assert not stmt
        return expr


class BoolVar(BoolExpression):
    """A boolean variable reference."""

    def __init__(self, var: PythonVar) -> None:
        self._var = var

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        bool_class = ctx.module.global_module.classes[PRIMITIVE_BOOL_TYPE]
        assert self._var.type is bool_class
        return self._var.ref()


class InhaleExhale(BoolExpression):
    """Inhale exhale expression."""

    def __init__(self, inhale: BoolExpression, exhale: BoolExpression) -> None:
        self._inhale = inhale
        self._exhale = exhale

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        inhale = self._inhale.translate(translator, ctx, position, info)
        exhale = self._exhale.translate(translator, ctx, position, info)
        return translator.viper.InhaleExhaleExp(
            inhale, exhale, position, info)


class BoolCall(BoolExpression, CallMixin):
    """A call to a boolean function.

    For example ``foo()`` where ``foo`` is defined as:

    .. code:: silver
        function foo(): Bool
    """

    def __init__(self, function: str, args: List['CallArg']) -> None:
        super().__init__(function, args, BOOL)  # Call CallMixin constructor.


class BigAnd(BoolExpression):
    """A conjunction of 0 or more elements.

    If it has zero elements, then it is equivalent to ``True``.
    """

    def __init__(self, conjuncts: List[BoolExpression]) -> None:
        self._conjuncts = conjuncts

    def is_empty(self) -> None:
        """Check if have any conjuncts."""
        return not self._conjuncts

    def is_always_true(self) -> bool:
        return self.is_empty()

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        if self.is_always_false():
            return translator.viper.FalseLit(position, info)
        conjuncts = [
            conjunct.translate(translator, ctx, position, info)
            for conjunct in self._conjuncts
            if not conjunct.is_always_true()]
        if not conjuncts:
            return translator.viper.TrueLit(position, info)
        else:
            and_operator = (
                lambda left, right:
                translator.viper.And(left, right, position, info))
            return join_expressions(and_operator, conjuncts)


class BigOr(BoolExpression):
    """A disjunction of 0 or more elements.

    If it has zero elements, then it is equivalent to ``False``.
    """

    def __init__(self, disjuncts: List[BoolExpression]) -> None:
        self._disjuncts = disjuncts

    def is_empty(self) -> None:
        """Check if have any elements."""
        return not self._disjuncts

    def is_always_false(self) -> bool:
        return self.is_empty()

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        if self.is_always_true():
            return translator.viper.TrueLit(position, info)
        disjuncts = [
            disjunct.translate(translator, ctx, position, info)
            for disjunct in self._disjuncts
            if not disjunct.is_always_false()]
        if not disjuncts:
            return translator.viper.FalseLit(position, info)
        else:
            or_operator = (
                lambda left, right:
                translator.viper.Or(left, right, position, info))
            return join_expressions(or_operator, disjuncts)


class BoolCondExp(BoolExpression):
    """A ternary operator with boolean result."""

    def __init__(self, condition: BoolExpression, then: BoolExpression,
                 els: BoolExpression) -> None:
        self._condition = condition
        self._then = then
        self._els = els

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        condition = self._condition.translate(translator, ctx, position, info)
        then = self._then.translate(translator, ctx, position, info)
        els = self._els.translate(translator, ctx, position, info)
        return translator.viper.CondExp(
            condition, then, els, position, info)


class Implies(BoolExpression):
    """Implication."""

    def __init__(self, condition: BoolExpression, value: Expression) -> None:
        self._condition = condition
        self._value = value

    def is_always_true(self) -> bool:
        return (self._condition.is_always_false() or
                self._value.is_always_true())

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        value = self._value.translate(translator, ctx, position, info)
        if self._condition.is_always_true():
            return value
        else:
            condition = self._condition.translate(
                translator, ctx, position, info)
            return translator.viper.Implies(
                condition, value, position, info)


class EqCmp(BoolExpression):
    """Equality comparison."""

    def __init__(self, left: Expression, right: Expression) -> None:
        self._left = left
        self._right = right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.EqCmp(
            left, right, position, info)


Expression.EqCmp = EqCmp


class NeCmp(BoolExpression):
    """Inequality comparison."""

    def __init__(self, left: Expression, right: Expression) -> None:
        self._left = left
        self._right = right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.NeCmp(
            left, right, position, info)


Expression.NeCmp = NeCmp



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast
from typing import List

from nagini_translation.lib.constants import ARBITRARY_BOOL_FUNC
from nagini_translation.lib.program_nodes import PythonClass, PythonTryBlock, PythonVar
from nagini_translation.lib.typedefs import Expr, Position, Seqn, Stmt, Var, VarDecl
from nagini_translation.lib.util import flatten, get_body_indices, get_surrounding_try_blocks
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.statement import StatementTranslator


class SIFStatementTranslator(StatementTranslator):
    """
    Extended AST Version of the StatementTranslator
    """

    def get_error_var(self, stmt: ast.AST, ctx: Context) -> PythonVar:
        return ctx.current_function.error_var

    def _translate_return(self, node: ast.Return, ctx: Context) -> List[Stmt]:
        if node.value:
            rhs_stmt, rhs = self.translate_expr(node.value, ctx)
        else:
            rhs_stmt, rhs = [], self.viper.NullLit(self.no_position(ctx),
                                                   self.no_info(ctx))
        pos = self.to_position(node, ctx)
        info = self.no_info(ctx)
        if ctx.result_var:
            rvar = ctx.result_var.ref(node, ctx)
        else:
            rvar = None
        ret_stmt = self.viper.Return(rhs, rvar, pos, info)
        return rhs_stmt + [ret_stmt]

    def translate_stmt_Return(self, node: ast.Return, ctx: Context) -> List[Stmt]:
        return self._translate_return(node, ctx)

    def translate_stmt_Break(self, node: ast.Break, ctx: Context) -> List[Stmt]:
        return [self.viper.Break(self.to_position(node, ctx), self.no_info(ctx))]

    def translate_stmt_Continue(self, node: ast.Continue, ctx: Context) -> List[Stmt]:
        return [self.viper.Continue(self.to_position(node, ctx), self.no_info(ctx))]

    def _while_postamble(self, node: ast.While, post_label: str,
                         ctx: Context) -> List[Stmt]:
        return self._set_result_none(ctx)

    def _translate_while_body(self, node: ast.While, ctx: Context,
                              end_label: str) -> List[Stmt]:
        start, end = get_body_indices(node.body)
        body = flatten(
            [self.translate_stmt(stmt, ctx) for stmt in node.body[start:end]])
        return body

    def _create_try_handlers(self, node: ast.Try, try_block: PythonTryBlock,
                             ctx: Context) -> List['silver.sif.SIFExceptionHandler']:
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)
        catch_blocks = []
        error_var = self.get_error_var(node, ctx)
        for handler in try_block.handlers:
            error_type_check = self.type_check(
                error_var.ref(), handler.exception, self.to_position(handler.node, ctx),
                ctx, inhale_exhale=False)
            handler_body = []
            if handler.exception_name:
                ctx.var_aliases[handler.exception_name] = error_var

                error_var.type = handler.exception
                handler_body.append(self.set_var_defined(error_var, no_pos, no_info))
            handler_body += flatten([self.translate_stmt(stmt, ctx)
                                     for stmt in handler.body])
            handler_body_seqn = self.viper.Seqn(handler_body, no_pos, no_info)
            catch_blocks.append(self.viper.SIFExceptionHandler(
                error_var.ref(), error_type_check, handler_body_seqn))
            if handler.exception_name:
                del ctx.var_aliases[handler.exception_name]
        return catch_blocks

    def translate_stmt_Try(self, node: ast.Try, ctx: Context) -> List[Stmt]:
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)
        try_block = self._get_try_block(node, ctx)
        assert try_block
        body = flatten([self.translate_stmt(stmt, ctx) for stmt in node.body])
        if try_block.else_block:
            else_block = flatten([self.translate_stmt(stmt, ctx)
                                  for stmt in try_block.else_block.body])
            else_block = self.viper.Seqn(else_block, no_pos, no_info)
        else:
            else_block = None
        if try_block.finally_block:
            finally_block = flatten([self.translate_stmt(stmt, ctx)
                                     for stmt in try_block.finally_block])
            finally_block = self.viper.Seqn(finally_block, no_pos, no_info)
        else:
            finally_block = None
        catch_blocks = self._create_try_handlers(node, try_block, ctx)
        return [self.viper.Try(self.viper.Seqn(body, no_pos, no_info),
                               catch_blocks, else_block, finally_block,
                               self.to_position(node, ctx), no_info)]

    def translate_stmt_Raise(self, node: ast.Raise, ctx: Context) -> List[Stmt]:
        err_var = self.get_error_var(node, ctx).ref()
        stmts = self._translate_stmt_raise_create(node, err_var, ctx)
        return stmts[:-1] + [
            self.viper.Raise(stmts[-1], self.to_position(node, ctx), self.no_info(ctx))
            ]

    def _translate_With_body(self,
                             try_block: PythonTryBlock,
                             enter_res: PythonVar,
                             ctx: Context) -> List[Stmt]:
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)
        if try_block.with_item.optional_vars:
            as_expr = try_block.with_item.optional_vars
            as_var = ctx.current_function.get_variable(as_expr.id)
            enter_assign = self.viper.LocalVarAssign(as_var.ref(as_expr, ctx),
                                                     enter_res.ref(),
                                                     self.to_position(as_expr,
                                                                      ctx),
                                                     no_info)
            define_var = self.set_var_defined(as_var, no_pos, no_info)
            body = [enter_assign, define_var]
        else:
            body = []
        body += flatten([self.translate_stmt(stmt, ctx) for stmt in try_block.node.body])
        return body

    def _create_With_ecx_var(self, ctx: Context) -> (VarDecl, Var):
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)
        no_exc_name = ctx.current_function.get_fresh_name('exc')
        no_exc_decl = self.viper.LocalVarDecl(no_exc_name, self.viper.Bool, no_pos,
                                              no_info)
        no_exc = self.viper.LocalVar(no_exc_name, self.viper.Bool, no_pos, no_info)
        return no_exc_decl, no_exc

    def _create_With_exception_handler_type_check(self, try_block: PythonTryBlock,
                                                  ctx: Context) -> Expr:
        exception_class = ctx.module.global_module.classes['Exception']
        error_var = self.get_error_var(try_block.node, ctx).ref()
        return self.type_check(error_var, exception_class,
                               self.no_position(ctx), ctx, inhale_exhale=False)

    def _create_With_exception_handler(self, try_block: PythonTryBlock,
                                       exit_res: PythonVar, no_exc: 'viper.ast.LocalVar',
                                       ctx: Context) -> Seqn:
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)
        ctx_type = self.get_type(try_block.with_item.context_expr, ctx)
        error_var = self.get_error_var(try_block.node, ctx).ref()
        err_type_arg = self.type_factory.typeof(error_var, ctx)

        tb_class = ctx.module.global_module.classes['traceback']
        traceback_var = ctx.actual_function.create_variable('tb', tb_class,
                                                            self.translator)
        tb_type = self.type_check(traceback_var.ref(), tb_class, no_pos, ctx)
        inhale_types = self.viper.Inhale(tb_type, no_pos, no_info)

        exit_call_err = self.get_method_call(ctx_type, '__exit__',
                                             [try_block.with_var.ref(), err_type_arg,
                                              error_var,
                                              traceback_var.ref()],
                                             [ctx_type, None, None, None],
                                             [exit_res.ref()],
                                             try_block.with_item.context_expr, ctx)

        exit_method_node = ctx_type.get_method('__exit__').node.returns
        exit_result = self.to_bool(exit_res.ref(), ctx, exit_method_node)
        return self.viper.Seqn([
            self.viper.LocalVarAssign(
                no_exc, self.viper.FalseLit(no_pos, no_info), no_pos, no_info),
            inhale_types
            ] + exit_call_err + [
                self.viper.If(self.viper.Not(exit_result, no_pos, no_info),
                              self.viper.Raise(None, no_pos, no_info),
                              self.viper.Skip(), no_pos, no_info),
            ], no_pos, no_info)

    def _translate_With_inner_try(self,
                                  try_block: PythonTryBlock,
                                  enter_res: PythonVar,
                                  exception_handler,
                                  ctx: Context) -> 'viper.silver.sif.SIFTryCatchStmt':
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)
        body = self._translate_With_body(try_block, enter_res, ctx)
        return self.viper.Try(self.viper.Seqn(body, no_pos, no_info),
                              [exception_handler], None, None, no_pos, no_info)

    def translate_With_finally_block(self, try_block: PythonTryBlock, exit_res: PythonVar,
                                     no_exc: Var, ctx: Context):
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)
        null = self.viper.NullLit(no_pos, no_info)
        exit_call_no_err = self.get_method_call(try_block.with_var.type, '__exit__',
                                                [try_block.with_var.ref(),
                                                 self.type_factory.typeof(
                                                     self.viper.NullLit(no_pos, no_info),
                                                     ctx),
                                                 null, null],
                                                [try_block.with_var.type, None, None, None],
                                                [exit_res.ref()],
                                                try_block.with_item.context_expr, ctx)
        return self.viper.If(no_exc,
                             self.viper.Seqn(exit_call_no_err, no_pos, no_info),
                             self.viper.Skip(), no_pos, no_info)

    def translate_stmt_With(self, node: ast.With, ctx: Context) -> List[Stmt]:
        """
        Transform a with statement into extended AST. See translation from PEP 343:

        with EXPR as VAR:
            BLOCK

        is translated to

        mgr = (EXPR)
        exit = type(mgr).__exit__  # Not calling it yet
        value = type(mgr).__enter__(mgr)
        exc = True
        try:
            try:
                VAR = value  # Only if "as VAR" is present
                BLOCK
            except:
                # The exceptional case is handled here
                exc = False
                if not exit(mgr, *sys.exc_info()):
                    raise
                # The exception is swallowed if exit() returns true
        finally:
            # The normal and non-local-goto cases are handled here
            if exc:
                exit(mgr, None, None, None)
        """
        try_block = self._get_try_block(node, ctx)
        assert try_block
        pos = self.to_position(node, ctx)
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)
        # Get context mgr
        ctx_stmt, ctx_mgr = self.translate_expr(try_block.with_item.context_expr, ctx)
        ctx_type = self.get_type(try_block.with_item.context_expr, ctx)
        # Create and assign temp variable
        with_ctx = ctx.current_function.create_variable('with_ctx', ctx_type, self.translator)
        try_block.with_var = with_ctx
        ctx_assign = self.viper.LocalVarAssign(with_ctx.ref(), ctx_mgr, no_pos, no_info)
        # Call __enter__ method
        enter_method_type = ctx_type.get_method('__enter__').type
        enter_res = ctx.current_function.create_variable('enter_res',
                                                         enter_method_type,
                                                         self.translator)
        enter_call = self.get_method_call(ctx_type, '__enter__',
                                          [with_ctx.ref()],
                                          [ctx_type],
                                          [enter_res.ref(node, ctx)], node, ctx)
        # Create ecx variable for checking if we raise an exception (see docstring)
        no_exc_decl, no_exc = self._create_With_ecx_var(ctx)
        exc_assign = self.viper.LocalVarAssign(
            no_exc, self.viper.TrueLit(no_pos, no_info), no_pos, no_info)
        # Create variable to take result of __exit__
        exit_res = ctx.current_function.create_variable(
            'exit_res', ctx_type.get_method('__exit__').type, self.translator)
        # Create inner try block
        handler = self._create_With_exception_handler(try_block, exit_res, no_exc, ctx)
        error_type_check = self._create_With_exception_handler_type_check(try_block, ctx)
        exception_handler = self.viper.SIFExceptionHandler(
            self.get_error_var(node, ctx).ref(), error_type_check, handler)
        inner_try = self._translate_With_inner_try(try_block, enter_res, exception_handler,
                                                   ctx)
        # Create outer try block
        finally_block = self.translate_With_finally_block(try_block, exit_res, no_exc, ctx)
        outer_try = self.viper.Try(self.viper.Seqn([inner_try], no_pos, no_info), [], None,
                                   self.viper.Seqn([finally_block], no_pos, no_info), pos,
                                   no_info)

        return [self.viper.Seqn(ctx_stmt + [ctx_assign] +
                                enter_call + [exc_assign, outer_try],
                                pos, no_info, locals=[no_exc_decl])]



from typing import Callable, Tuple, Any


class Thread:
    """
    A stub for threads with a similar interface as threading.Thread.
    Clients must give the constructor a target method; subclassing and overriding
    the run method is currently not supported.
    """
    def __init__(self, group: object=None, target: Callable=None, name: str=None,
                 args: Tuple=None, kwargs: object=None, daemon: bool=None) -> None :
        pass

    def start(self, *method_list : Callable) -> None:
        """
        Starts this thread; requires a permission to start it.
        method_list must include a reference to the actual target method of this thread
        that was passed to it upon creation.
        """
        pass

    def join(self, *method_list : Callable) -> None:
        """
        Joins this thread, requires a permission to do so. If a ThreadPost permission is
        held and the thread's actual target method is included in method_list, this will
        inhale the thread's postcondition.
        """
        pass


def getMethod(t: Thread) -> Callable:
    """Return the target method of a thread."""
    pass


def getArg(t: Thread, i: int) -> object:
    """
    Return the ith element passed to the thread constructor in the ``args`` parameter,
    unless the thread target passed to the constructor contained a receiver object, in
    which case this will return the (i+1)th argument element from the ``args`` tuple
    and the receiver object is treated as the argument with index zero.

    t = Thread(target=foo, args=(x, y, z))  # getArg(t, 0) == x, getArg(t, 2) == z
    t = thread(target=o.bar, args=(x, y, z))  # getArg(t, 0) == o, getArg(t, 1) == x
    """
    pass


def getOld(t: Thread, o: object) -> object:
    """
    Return the value of expression o when thread t was started.
    o must be an expression mentioned in an old-expression in the postcondition of the
    thread's target method, in which references to parameter i of the target method are
    replaces by arg(i) (see below).
    """
    pass


def getARP(t: Thread) -> float:
    """
    Returns the permission value of abstract read permissions passed to thread t.
    """
    pass


def arg(i: int) -> Any:
    """Used to reference method parameters in getOld() calls."""
    pass


def Joinable(t: Thread) -> bool:
    """
    Represents that thread t is joinable.
    Threads may be joined if they promise to terminate.
    """
    pass


def MayStart(t: Thread) -> bool:
    """A permission to start a newly created thread. Can only be used once."""
    pass


def ThreadPost(t: Thread) -> bool:
    """
    A permission to get the postcondition of thread t after joining it.
    If only a fractional amount of permission to ThreadPost(t) is held when joining,
    only that fractional amount of the postcondition will be inhaled.
    Holding some permission to ThreadPost for a thread implies that the thread
    is joinable; it is not necessary to write ThreadPost and Joinable.
    """
    pass



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Code for constructing Silver while node with obligation stuff."""


import ast
from typing import List, Union

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.config import obligation_config
from nagini_translation.lib.context import Context
from nagini_translation.lib.errors import rules
from nagini_translation.lib.typedefs import (
    Expr,
    Stmt,
    VarDecl,
)
from nagini_translation.sif.lib.viper_ast_extended import ViperASTExtended
from nagini_translation.translators.obligation.manager import (
    ObligationManager,
)
from nagini_translation.translators.obligation.node_constructor import (
    StatementNodeConstructorBase,
)
from nagini_translation.translators.obligation.obligation_info import (
    PythonLoopObligationInfo,
)
from nagini_translation.translators.obligation.utils import (
    bound_obligations,
)


class ObligationLoop:
    """Info for generating Silver ``While`` AST node."""

    def __init__(
            self, condition: Expr, invariants: List[Expr],
            local_vars: List[VarDecl], body: List[Stmt]) -> None:
        self.condition = condition
        self.invariants = invariants
        self.local_vars = local_vars
        self.body = body

    def prepend_invariant(self, invariant: Expr) -> None:
        """Prepend ``invariant`` to the invariants list."""
        self.invariants.insert(0, invariant)

    def append_invariants(self, invariants: List[Expr]) -> None:
        """Append ``invariants`` to the invariants list."""
        self.invariants.extend(invariants)

    def prepend_body(self, statements: List[Stmt]) -> None:
        """Prepend ``statements`` to body."""
        self.body[0:0] = statements

    def append_body(self, statement: Stmt) -> None:
        """Append ``statement`` to body."""
        self.body.append(statement)


class ObligationLoopNodeConstructor(StatementNodeConstructorBase):
    """A class that creates a while loop node with obligation stuff."""

    def __init__(
            self, obligation_loop: ObligationLoop,
            node: Union[ast.While, ast.For],
            translator: 'AbstractTranslator', ctx: Context,
            obligation_manager: ObligationManager) -> None:
        position = translator.to_position(node, ctx)
        info = translator.no_info(ctx)
        super().__init__(
            translator, ctx, obligation_manager, position, info, node)
        self._obligation_loop = obligation_loop
        self._node = node

    def get_statements(self) -> List[Stmt]:
        """Get all generated statements."""
        return self._statements

    def construct_loop(self) -> None:
        """Construct statements to perform a loop."""
        self._add_current_wait_level()
        self._add_additional_invariants()
        self._add_leak_check()
        self._set_up_measures()
        self._bound_obligations()
        self._save_must_terminate_amount(
            self._loop_obligation_info.original_must_terminate_var)
        self._save_loop_termination()
        self._set_loop_check_before()
        self._set_loop_check_after_body()
        self._check_loop_preserves_termination()
        self._prepend_loop_body()
        self._add_loop()
        self._add_additional_statements_after_loop()
        self._reset_must_terminate(
            self._loop_obligation_info.original_must_terminate_var)

    def _add_current_wait_level(self) -> None:
        """Inhale assumptions about current wait-level variable."""
        if obligation_config.disable_waitlevel_check:
            return
        context_info = self._ctx.obligation_context.get_surrounding_loop_info()
        if context_info:
            context_residue_level = context_info.residue_level
        else:
            method_info = self._ctx.current_function.obligation_info
            context_residue_level = method_info.residue_level
        invariant = self._translator.initialize_current_wait_level(
            sil.PermVar(self._loop_obligation_info.residue_level),
            sil.PermVar(context_residue_level),
            self._ctx)
        translated = invariant.translate(
            self._translator, self._ctx, self._position, self._info)
        self._obligation_loop.prepend_invariant(translated)

    def _add_additional_invariants(self) -> None:
        """Add additional invariants from the obligation info."""
        self._obligation_loop.append_invariants(
            self._loop_obligation_info.get_additional_invariants())

    def _add_leak_check(self) -> None:
        """Add leak checks to invariant."""
        reference_name = self._ctx.actual_function.get_fresh_name('_r')
        leak_check = self._obligation_manager.create_leak_check(reference_name)
        loop_check_before = sil.BoolVar(
            self._loop_obligation_info.loop_check_before_var)
        termination_flag = sil.BoolVar(
            self._loop_obligation_info.termination_flag_var)

        if not obligation_config.disable_loop_context_leak_check:
            must_terminate = self._obligation_manager.must_terminate_obligation
            predicate = must_terminate.create_predicate_access(
                self._obligation_info.current_thread_var)
            termination_leak_check = sil.CurrentPerm(predicate) == sil.NoPerm()
            loop_cond = self._loop_obligation_info.construct_loop_condition()
            before_loop_leak_check = sil.InhaleExhale(
                sil.TrueLit(),
                sil.Implies(
                    loop_check_before,
                    sil.BigOr([
                        termination_flag,
                        sil.Not(loop_cond),
                        sil.BigAnd([termination_leak_check, leak_check])
                    ])
                )
            )
            info = self._to_info('Leak check for context.')
            position = self._to_position(
                conversion_rules=rules.OBLIGATION_LOOP_CONTEXT_LEAK_CHECK_FAIL)
            self._obligation_loop.append_invariants([
                before_loop_leak_check.translate(
                    self._translator, self._ctx, position, info)])

        if not obligation_config.disable_loop_body_leak_check:
            body_leak_check = sil.InhaleExhale(
                sil.TrueLit(),
                sil.Implies(sil.Not(loop_check_before), leak_check))
            info = self._to_info('Leak check for loop body.')
            position = self._to_position(
                conversion_rules=rules.OBLIGATION_LOOP_BODY_LEAK_CHECK_FAIL)
            self._obligation_loop.append_invariants([
                body_leak_check.translate(
                    self._translator, self._ctx, position, info)])

    def _set_up_measures(self) -> None:
        """Create and initialize loop's measure map."""
        if obligation_config.disable_measures:
            return
        # Set up measures.
        loop_measure_map = self._loop_obligation_info.loop_measure_map
        instances = self._loop_obligation_info.get_all_instances()
        statements = loop_measure_map.initialize(
            instances, self._translator, self._ctx)
        self._obligation_loop.prepend_body(statements)

    def _bound_obligations(self) -> None:
        """Convert all unbounded obligations to bounded ones."""
        statements = bound_obligations(
            self._loop_obligation_info.get_all_instances(),
            self._translator, self._ctx, self._position, self._info)
        self._obligation_loop.prepend_body(statements)

    def _save_loop_termination(self) -> None:
        """Save if loop promises to terminate into a variable."""
        if obligation_config.disable_termination_check:
            return
        assign = sil.Assign(
            self._loop_obligation_info.termination_flag_var,
            self._loop_obligation_info.create_termination_check(True))
        info = self._to_info('Save loop termination promise.')
        self._append_statement(assign, info=info)

    def _set_loop_check_before(self) -> None:
        """Set the variable indicating that we are before loop."""
        if obligation_config.disable_all:
            return
        assign = sil.Assign(
            self._loop_obligation_info.loop_check_before_var,
            sil.TrueLit())
        info = self._to_info('We are before loop.')
        self._append_statement(assign, info=info)

    def _set_loop_check_after_body(self) -> None:
        """Set the variable indicating that we are after loop body."""
        if obligation_config.disable_all:
            return
        assign = sil.Assign(
            self._loop_obligation_info.loop_check_before_var,
            sil.FalseLit())
        info = self._to_info('We are after loop body.')
        statement = assign.translate(
            self._translator, self._ctx, self._position, info)
        self._obligation_loop.append_body(statement)

    def _check_loop_preserves_termination(self) -> None:
        """Check that loop keeps the promise to terminate."""
        if obligation_config.disable_termination_check:
            return
        check = sil.Implies(
            sil.BoolVar(self._loop_obligation_info.termination_flag_var),
            sil.BigOr([
                sil.Not(self._loop_obligation_info.construct_loop_condition()),
                self._loop_obligation_info.create_termination_check(False)
            ])
        )
        assertion = sil.Assert(check)
        position = self._to_position(
            conversion_rules=rules.OBLIGATION_LOOP_TERMINATION_PROMISE_FAIL)
        comment = 'Check if loop continues to terminate.'
        if isinstance(self._viper, ViperASTExtended):
            info = self._viper.SIFInfo([comment], continue_unaware=True)
        else:
            info = self._to_info(comment)
        statement = assertion.translate(
            self._translator, self._ctx, position, info)
        self._obligation_loop.append_body(statement)

    def _prepend_loop_body(self) -> None:
        """Add additional statements before loop from the obligation info."""
        self._obligation_loop.prepend_body(
            self._loop_obligation_info.get_prepend_body())

    def _add_loop(self) -> None:
        """Add the actual loop node."""
        body_block = self._translator.translate_block(
            self._obligation_loop.body, self._position, self._info)
        loop = self._viper.While(
            self._obligation_loop.condition, self._obligation_loop.invariants,
            self._obligation_loop.local_vars, body_block,
            self._position, self._info)
        self._statements.append(loop)

    def _add_additional_statements_after_loop(self) -> None:
        """Add additional statements after loop from the obligation info."""
        self._statements.extend(
            self._loop_obligation_info.get_after_loop())

    @property
    def _loop_obligation_info(self) -> PythonLoopObligationInfo:
        return self._ctx.obligation_context.current_loop_info



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Error handling state is stored in singleton ``manager``."""


from collections import namedtuple
from uuid import uuid1

from typing import Any, List, Optional

from nagini_translation.lib.errors.wrappers import Error
from nagini_translation.lib.errors.rules import Rules
from nagini_translation.lib.jvmaccess import JVM


Item = namedtuple('Item', 'node vias reason_string')


class ErrorManager:
    """A singleton object that stores the state needed for error handling."""

    def __init__(self) -> None:
        self._items = {}                # type: Dict[str, Item]
        self._conversion_rules = {}     # type: Dict[str, Rules]

    def add_error_information(
            self, node: 'ast.Node', vias: List[Any], reason_string: str,
            conversion_rules: Rules = None) -> str:
        """Add error information to state."""
        item_id = str(uuid1())
        assert item_id not in self._items
        self._items[item_id] = Item(node, vias, reason_string)
        if conversion_rules is not None:
            self._conversion_rules[item_id] = conversion_rules
        return item_id

    def clear(self) -> None:
        """Clear all state."""
        self._items.clear()
        self._conversion_rules.clear()

    def convert(
            self,
            errors: List['AbstractVerificationError'],
            jvm: Optional[JVM]) -> List[Error]:
        """Convert Viper errors into Nagini errors.

        It does that by wrapping in ``Error`` subclasses.
        """
        new_errors = [
            self._convert_error(error, jvm)
            for error in errors
        ]
        return new_errors

    def get_vias(self, node_id: str) -> List[Any]:
        """Get via information for the given ``node_id``."""
        item = self._items[node_id]
        return item.vias

    def _get_item(self, pos: 'ast.AbstractSourcePosition') -> Optional[Item]:
        if hasattr(pos, 'id'):
            node_id = pos.id()
            return self._items[node_id]
        return None

    def _get_conversion_rules(
            self, position: 'ast.AbstractSourcePosition') -> Optional[Rules]:
        if hasattr(position, 'id'):
            node_id = position.id()
            if node_id in self._conversion_rules:
                return self._conversion_rules[node_id]
        return None

    def _try_get_rules_workaround(
            self, node: 'ast.Node', jvm: Optional[JVM]) -> Optional[Rules]:
        """Try to extract rules out of ``node``.

        Due to optimizations, Silicon sometimes returns not the correct
        offending node, but an And that contains it. This method tries
        to work around this problem.

        .. todo::

            In the long term we should discuss with Malte how to solve
            this problem properly.
        """
        rules = self._get_conversion_rules(node.pos())
        if rules or not jvm:
            return rules
        if (isinstance(node, jvm.viper.silver.ast.And) or
                isinstance(node, jvm.viper.silver.ast.Implies)):
            return (self._get_conversion_rules(node.left().pos()) or
                    self._get_conversion_rules(node.right().pos()) or
                    self._try_get_rules_workaround(node.left(), jvm) or
                    self._try_get_rules_workaround(node.right(), jvm))
        return

    def transformError(self, error: 'AbstractVerificationError') -> 'AbstractVerificationError':
        """ Transform silver error to a fixpoint. """
        old_error = None
        while old_error != error:
            old_error = error
            error = error.transformedError()
        return error

    def _convert_error(
            self, error: 'AbstractVerificationError',
            jvm: Optional[JVM]) -> Error:
        error = self.transformError(error)
        reason_pos = error.reason().offendingNode().pos()
        reason_item = self._get_item(reason_pos)
        position = error.pos()
        rules = self._try_get_rules_workaround(
            error.offendingNode(), jvm)
        if rules is None:
            rules = self._try_get_rules_workaround(
                error.reason().offendingNode(), jvm)
        if rules is None:
            rules = {}
        error_item = self._get_item(position)
        if error_item:
            return Error(error, rules, reason_item, error_item.node,
                         error_item.vias)
        else:
            return Error(error, rules, reason_item)


manager = ErrorManager()     # pylint: disable=invalid-name



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Utility functions."""


from typing import List

from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import PythonMethod
from nagini_translation.lib.typedefs import (
    Info,
    Position,
    Stmt,
)


def find_method_by_sil_name(ctx: Context, sil_name: str) -> PythonMethod:
    """Find Python method from the global module based on its Silver name."""
    module = ctx.module.global_module
    methods = module.methods
    for method in methods.values():
        if method.sil_name == sil_name:
            return method
    classes = module.classes
    for cls in classes.values():
        for method in cls.methods.values():
            if method.sil_name == sil_name:
                return method
    return None


def bound_obligations(
        instances: List['ObligationInstance'],
        translator: 'AbstractTranslator', ctx: Context,
        position: Position, info: Info) -> List[Stmt]:
    """Construct statements for bounding fresh obligations."""
    statements = []
    for instance in instances:
        if not instance.obligation_instance.is_fresh():
            continue
        statement = instance.obligation_instance.get_obligation_bound(ctx)
        statements.append(statement)
    translated_statements = [
        statement.translate(translator, ctx, position, info)
        for statement in statements]
    return translated_statements



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast
import astunparse

from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    TypeVar,
)


T = TypeVar('T')
V = TypeVar('V')


def flatten(lists: List[List[T]]) -> List[T]:
    """
    Flattens a list of lists into a flat list
    """
    return [item for sublist in lists for item in sublist]


def flatten_dict(dicts: List[Dict[T, List[V]]],
                 defaults: List[T]=[]) -> Dict[T, List[V]]:
    """
    Flattens a dict of lists, i.e., concatenates all lists for the same keys.
    """
    result = {}
    for key in defaults:
        result[key] = []
    for d in dicts:
        for key, value in d.items():
            if key in result:
                result[key].extend(value)
            else:
                result[key] = value
    return result


class UnsupportedException(Exception):
    """
    Exception that is thrown when attempting to translate a Python element not
    currently supported
    """

    def __init__(self, ast_element: ast.AST, desc=""):
        self.node = ast_element
        super().__init__(desc)


class InvalidProgramException(Exception):
    """
    Signals that the input program is invalid and cannot be translated
    """

    def __init__(self, node: ast.AST, code: str, message: str = None):
        self.node = node
        self.code = code
        self.message = message


class ConsistencyException(Exception):
    """
    Exception reporting that the translated AST has a consistency error
    """

    def __init__(self, message: str = None) -> None:
        self.message = message


class AssignCollector(ast.NodeVisitor):
    """
    Collects all assignment targets within a given (partial) AST.
    """
    def __init__(self):
        self.assigned_vars = {}

    def visit_Assign(self, node: ast.Assign) -> None:
        for target in node.targets:
            self._track_assign(target)

    def _track_assign(self, target: ast.AST) -> None:
        if isinstance(target, ast.Tuple):
            actual_targets = target.elts
        else:
            actual_targets = [target]
        for actual in actual_targets:
            if isinstance(actual, ast.Name):
                self.assigned_vars[actual.id] = actual

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        self._track_assign(node.target)


def get_func_name(stmt: ast.AST) -> Optional[str]:
    """
    Checks if stmt is a function call and returns its name if it is, None
    otherwise.
    """
    if isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Call):
        call = stmt.value
    elif isinstance(stmt, ast.Call):
        call = stmt
    else:
        return None
    if isinstance(call.func, ast.Name):
        return call.func.id
    elif isinstance(call.func, ast.Attribute):
        return call.func.attr
    else:
        raise UnsupportedException(stmt)


def get_column(node: ast.AST) -> Optional[int]:
    return node.col_offset if hasattr(node, 'col_offset') else None


def contains_stmt(container: Any, contained: ast.AST) -> bool:
    """
    Checks if 'contained' is a part of the partial AST
    whose root is 'container'.
    """
    if container is contained:
        return True
    if isinstance(container, list):
        for stmt in container:
            if contains_stmt(stmt, contained):
                return True
        return False
    elif isinstance(container, ast.AST):
        for field in container._fields:
            if contains_stmt(getattr(container, field), contained):
                return True
        return False
    else:
        return False


def get_surrounding_try_blocks(try_blocks: List['PythonTryBlock'],
                               stmt: ast.AST) -> List['PythonTryBlock']:
    """
    Finds the try blocks in try_blocks that protect the statement stmt.
    """
    def rank(b: 'PythonTryBlock', blocks: List['PythonTryBlock']) -> int:
        result = 0
        for b2 in blocks:
            if contains_stmt(b2.protected_region, b.node):
                result += 1
        return -result
    tb = try_blocks
    blocks = [b for b in tb if contains_stmt(b.protected_region, stmt)]
    inner_to_outer = sorted(blocks,key=lambda b: rank(b, blocks))
    return inner_to_outer


def is_invariant(stmt: ast.AST) -> bool:
    return get_func_name(stmt) == 'Invariant'


def is_pre(stmt: ast.AST) -> bool:
    return get_func_name(stmt) == 'Requires'


def is_post(stmt: ast.AST) -> bool:
    return get_func_name(stmt) == 'Ensures'


def is_exception_decl(stmt: ast.AST) -> bool:
    return get_func_name(stmt) == 'Exsures'


def is_docstring(stmt: ast.AST) -> bool:
    """Return True if statement is a docstring."""
    if (isinstance(stmt, ast.Expr) and
            isinstance(stmt.value, ast.Str)):
        return True
    else:
        return False


def is_io_existential(stmt: ast.AST) -> bool:
    """Return True if statement is a definition of IOExists."""
    if (isinstance(stmt, ast.Expr) and
            isinstance(stmt.value, ast.Call) and
            isinstance(stmt.value.func, ast.Call) and
            isinstance(stmt.value.func.func, ast.Name) and
            stmt.value.func.func.id.startswith('IOExists')):
        return True
    else:
        return False


def get_body_indices(statements: List[ast.AST]) -> Tuple[int, int]:
    """
    Returns the index of the first statement that is not a method
    or loop contract and the index after the last statement that is not a postcondition.

    .. note::

        In the case when a method has only a contract, both returned
        indices are equal to ``len(statements)``.
    """
    start_index = 0
    try:
        while is_docstring(statements[start_index]):
            start_index += 1
        while is_io_existential(statements[start_index]):
            start_index += 1
        while is_invariant(statements[start_index]):
            start_index += 1
        while is_pre(statements[start_index]):
            start_index += 1
        while is_post(statements[start_index]):
            start_index += 1
        while is_exception_decl(statements[start_index]):
            start_index += 1
    except IndexError:
        # This exception means that the method/loop has only a contract.
        pass
    end_index = len(statements)
    if start_index != end_index:
        while end_index > 0 and is_post(statements[end_index - 1]):
            end_index -= 1
    return start_index, end_index


def find_loop_for_previous(node: ast.AST, name: str) -> ast.For:
    """
    In a for loop like::

        for x in xs:

    ``Previous(x)`` refers to the objects processed in previous iterations.
    Given the Previous-call-node, this function returns the for loop to whose
    previous iterations the node refers.
    """
    if isinstance(node, ast.For):
        if isinstance(node.target, ast.Name):
            if node.target.id == name:
                return node
        elif (isinstance(node.target, ast.Tuple) and node.target.elts and
                  isinstance(node.target.elts[0], ast.Name)):
            if node.target.elts[0].id == name:
                return node
    if not hasattr(node, '_parent') or not node._parent:
        return None
    return find_loop_for_previous(node._parent, name)


def get_parent_of_type(node: ast.AST, typ: type) -> ast.AST:
    parent = node._parent
    while not isinstance(parent, ast.Module):
        if isinstance(parent, typ):
            return parent
        parent = parent._parent
    return None


def join_expressions(operator: Callable[[T, T], T],
                     expressions: List[T]) -> T:
    """
    Joins expressions with ``operator``.

    This function joins expressions backwards (the last two expressions
    are most nested) in order to avoid Silicon issue
    `241 <https://bitbucket.org/viperproject/silicon/issues/241>`_.
    """
    first = expressions[-1]

    def new_op(first, second, third):
        return operator(second, third)
    return join_three_expressions(new_op, expressions, expressions, first)


def join_three_expressions(operator: Callable[[T, T, T], T],
                           expressions: List[T], bools: List[T], first: T) -> T:
    """
    Joins three expressions with ``operator`` in the same way as ``join_expressions``.
    """
    result = expressions[-1]
    for part_expr, part_bool in zip(reversed(expressions[:-1]),
                                    reversed(bools[:-1])):
        result = operator(part_expr, part_bool, result)
    return result


def construct_lambda_prefix(line: int, column: Optional[int]) -> str:
    """
    Creates an identifier for a lambda expression based on its line and column.
    """
    return 'lambda{0}_{1}'.format(line,
                                  'unknown' if column is None else column)


def pprint(node) -> str:
    """
    Pretty prints a Python AST node. When given a string, just returns it.
    """
    if not node:
        raise ValueError(node)
    if isinstance(node, str):
        return node
    if isinstance(node, ast.FunctionDef):
        # Mainly for debugging, whenever this happens it's almost certainly
        # wrong.
        raise ValueError(node)
    res = astunparse.unparse(node)
    res = res.replace('\n', '')
    return res


def get_target_name(node: ast.AST) -> str:
    """
    Returns the name of the function this node belongs to. If it's a call,
    that's the name of the call target, if it's a function, that function's
    name. For any other node, the name of the containing function,
    """
    if (not isinstance(node, ast.Call) and
            not isinstance(node, ast.FunctionDef)):
        node = get_containing_member(node)
    if isinstance(node, ast.FunctionDef):
        return node.name
    func = node.func
    if isinstance(func, ast.Name):
        func = func.id
    if isinstance(func, ast.Attribute):
        func = func.attr
    return func


def get_containing_member(node: ast.AST) -> Optional[ast.FunctionDef]:
    """
    Returns the function this node belongs to, if any.
    """
    member = node
    while not isinstance(member, ast.FunctionDef) and member is not None:
        if hasattr(member, '_parent'):
            member = member._parent
        else:
            member = None
    return member


def is_get_ghost_output(node: ast.Assign) -> bool:
    return (isinstance(node.value, ast.Call) and
            isinstance(node.value.func, ast.Name) and
            node.value.func.id == 'GetGhostOutput')


class OldExpressionCollector(ast.NodeVisitor):
    """A visitor that collects all expressions inside Old()-calls."""
    def __init__(self):
        self.expressions = []

    def visit_Call(self, node: ast.Call) -> None:
        if isinstance(node.func, ast.Name) and node.func.id == 'Old':
            assert len(node.args) == 1
            self.expressions.append(node.args[0])
        else:
            self.generic_visit(node)


class OldExpressionTransformer(ast.NodeTransformer):
    """
    A visitor that replaces references to specific names by calls to arg(i), where
    i is the index of the name (given in self.arg_names).
    """
    def __init__(self):
        self.arg_names = []

    def visit_Name(self, node: ast.Name):
        if node.id in self.arg_names:
            index = self.arg_names.index(node.id)
            return ast.Call(func=ast.Name(id='arg', ctx=ast.Load()),
                            args=[ast.Num(n=index)],
                            keywords=[])
        return node


class SingletonFreshName:
    """
    This class wraps the fresh name facility in scope by using it only when
    a new name is given. It is designed to store and retrieve the fresh name
    based on the original name given.
    """

    def __init__(self, scope: 'PythonScope') -> None:
        self._fresh_name_dict = {}
        self._scope = scope

    def __call__(self, name: str) -> str:
        """
        Returns a fresh name for a given name and scope if enquired for the
        first time, otherwise returns the previously given fresh name.
        """
        if name not in self._fresh_name_dict:
            self._fresh_name_dict[name] = self._scope.get_fresh_name(name)
        return self._fresh_name_dict[name]
